﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.Globalization;
using System.Reflection;
using System.Text;
using System.Threading;
using Produmex.Foundation;
using Produmex.Foundation.Data.DbClient;
using Produmex.Foundation.Data.Sbo;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions.Tables;
using Produmex.Foundation.Data.Sbo.DataObjects;
using Produmex.Foundation.Data.Sbo.Utilities;
using Produmex.Foundation.Data.SqlClient;
using Produmex.Foundation.Diagnostics;
using Produmex.Foundation.Messages;
using Produmex.Foundation.SlimScreen;
using Produmex.Foundation.SlimScreen.Interfaces;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions.KnownDataSets;
using Produmex.Foundation.Workflows;
using Produmex.Foundation.Workflows.Parameters;
using Produmex.Foundation.Wwf;
using Produmex.Foundation.Wwf.Sbo;
using Produmex.Foundation.Wwf.Sbo.LocalServices;
using Produmex.Sbo.Logex.Data;
using Produmex.Sbo.Logex.Data.BusinessObjects;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.DataObjects;
using Produmex.Sbo.Logex.Data.Providers;
using Produmex.Sbo.Logex.Gymna.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.ViewObjects.Definitions.Views;


// PmxWorkflowExecutionTypes.HOOK_FLOW
namespace WorkflowScript_AfterPickListPackedHookScript
{
    /// <summary>
    /// This is a hook script, meant to be customized on a "per project" basis
    /// </summary>
    public class WorkflowScript_AfterPickListPackedHookScript : WorkflowInstanceScriptBase
    {

     

        private static readonly ILog s_log = LogProvider.GetLogger(MethodInfo.GetCurrentMethod().DeclaringType);

        // Input parameters * do not change *
        public ReadOnlyBinder<CultureInfo> DefaultCultureInfo;
        public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;
        public ReadOnlyBinder<int> PickListDocEntry;
        public ReadWriteBinder<bool> DoAutomaticShipping;


        // Output parameters
        public ReadWriteBinder<Dictionary<string, object>> DeliveryData;

        // Sub-flows
        // <none>

        /// <summary>
        /// Gets or sets the delivery data.
        /// </summary>
        /// <value>The delivery data.</value>
        private Dictionary<string, object> deliveryData
        {
            get { return DeliveryData.Get(); }
            set { DeliveryData.Set(value); }
        }

        public const int DECIMALS = 6;
        public const string PICKLISTTYPE = "GLS";

        /// <summary>
        /// Initializes a new instance of the <see cref="WorkflowScript_AfterPickListPackedHookScript"/> class.
        /// </summary>
        /// <param name="parent">The parent.</param>
        /// <param name="factory">The factory.</param>
        public WorkflowScript_AfterPickListPackedHookScript(WorkflowInstanceBase parent, WorkflowInstanceFactory factory)
            : base(parent, factory)
        {
        }

        #region WorkflowInstanceScriptBase Members

        protected override void Execute()
        {
            // Parameters in scope
            Session session = GetScopeParameter("Session") as Session;
            ISboProviderService sboProviderService = GetScopeParameter("<WwfService>ISboService") as ISboProviderService;

            // 2016-02-25  KrisM  This hook-script will be called for each pick-list in the delivery,
            //                    but as we are returning the same UDF's only the first once returned will be kept!
            deliveryData = new Dictionary<string, object>();

            
            string initialErrorKey = null;
            Message msg = null;
            string query = null;
            int? numberofColli;
            double? totalWeight;
            double defaultQty = 0.0;
            DataSet dsTotalWeight = null;
            DataSet dsPicklistType = null;
            string picklistType = null;





        Step_EnterNumberOfColli:

            numberofColli = null;
            totalWeight = null;

            #region ENTER NUMBER OF COLLI

            query = BuildQuery.GetPicklistType( PickListDocEntry, DefaultCultureInfo );
            dsPicklistType = sboProviderService.RunView( false, null, null, query );
            if ( GetNumberOfRows( dsPicklistType ) > 0 )
            {
                picklistType = Convert.ToString(dsPicklistType.Tables[0].Rows[0][0]);
            }

            if ( picklistType == PICKLISTTYPE )
            {

                //ask for quantity
                session.ShowScreen( typeof( Produmex.Foundation.SlimScreen.Interfaces.IEnterQuantityScreen ),
                    DefaultCultureInfo.Get(), BuildParamCollection(
                        "InitialErrorKey", initialErrorKey,
                        "TitleKey", "MSG_TITLE_ENTER_NUMBER_OF_COLLI",
                        "InitialQuantity", 0,
                        "NumberOfDigitsAfterDecimal", 0,
                        "MaximumQuantity", -1
                       ) );
                msg = WaitForMessage();

                //clear data
                initialErrorKey = null;

                if ( msg.Name.EndsWith( ".QuantityEntered" ) )
                {
                    numberofColli = Convert.ToInt32( ExtractParameter<double>( msg.Parameters, "quantity" ) );
                    goto Step_EnterTotalWeight;
                }
                else
                {
                    goto Step_EnterNumberOfColli;
                }
            }

            #endregion

        Step_EnterTotalWeight:

            #region ENTER TOTAL WEIGHT

            if ( picklistType == PICKLISTTYPE )
            {
                query = BuildQuery.GetTotalWeight( PickListDocEntry.Get(), sboProviderService.GetDbTool() );
                dsTotalWeight = sboProviderService.RunView( false, null, null, query );
                if ( GetNumberOfRows( dsTotalWeight ) > 0 )
                {
                    defaultQty = Convert.ToDouble( dsTotalWeight.Tables[0].Rows[0][0] );
                }
                if ( defaultQty > 0 )
                {
                    session.ShowScreen( typeof( Produmex.Foundation.SlimScreen.Interfaces.IEnterQuantityScreen ),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "InitialErrorKey", initialErrorKey,
                            "TitleKey", "MSG_TITLE_ENTER_WEIGHT",
                            "InitialQuantity", defaultQty,
                            "NumberOfDigitsAfterDecimal", DECIMALS,
                            "MaximumQuantity", -1
                           ) );

                    msg = WaitForMessage();

                    if (msg.Name.EndsWith(".QuantityEntered"))
                    {
                        totalWeight = ExtractParameter<double>(msg.Parameters, "quantity");
                    }
                    else if (msg.Name.EndsWith(".BackRequested"))
                    {
                        goto Step_EnterNumberOfColli;
                    }
                    else
                    {//
                    }
                }
            }
            #endregion

            if (numberofColli.HasValue)
            {
                deliveryData.Add("U_PMX_G_COLLI", numberofColli.Value);
            }
            if (totalWeight.HasValue)
            {
                deliveryData.Add("U_PMX_G_WEIGHT", totalWeight.Value);
            }


        }


        #endregion

        #region HELPERS

        /// <summary>
        /// Gets the number of rows.
        /// </summary>
        /// <param name="ds">The ds.</param>
        /// <returns></returns>
        private int GetNumberOfRows(DataSet ds)
        {
            if (ds != null && ds.Tables.Count > 0)
            {
                return ds.Tables[0].Rows.Count;
            }
            return 0;
        }



        #endregion


        #region QUERY

        /// <summary>
        /// Static class grouping our native queries
        /// </summary>
        private class BuildQuery
        {
			public static string GetTotalWeight(int pickListDocEntry, DbTool dbTool)
			{
				/*
                
 SELECT SUM( "PMX_OITM_MEASUREMENTS"."SWeight1" * "OWGT"."WightInMG" * 1E-6 * "PMX_PLLI"."Quantity" ) AS "Total weight [kg]"
 FROM "PMX_PLLI"
 INNER JOIN "OITM" ON "OITM"."ItemCode" = "PMX_PLLI"."ItemCode"
 INNER JOIN "PMX_PLPL" ON "PMX_PLPL"."DocEntry" = "PMX_PLLI"."BaseEntry" AND "PMX_PLPL"."LineNum" = "PMX_PLLI"."BaseLine" AND "PMX_PLLI"."BaseType" = N'PMX_PLPH'
 INNER JOIN "PMX_PLPH" ON "PMX_PLPH"."DocEntry" = "PMX_PLPL"."DocEntry"
 LEFT JOIN "RDR1" ON "RDR1"."DocEntry" = "PMX_PLPL"."BaseEntry" AND "RDR1"."LineNum" = "PMX_PLPL"."BaseLine" AND "RDR1"."ObjType" = "PMX_PLPL"."BaseType"
 LEFT JOIN "INV1" ON "INV1"."DocEntry" = "PMX_PLPL"."BaseEntry" AND "INV1"."LineNum" = "PMX_PLPL"."BaseLine" AND "INV1"."ObjType" = "PMX_PLPL"."BaseType"
 LEFT JOIN "PMX_OITM_MEASUREMENTS" ON "PMX_OITM_MEASUREMENTS"."ItemCode" = "OITM"."ItemCode" AND ISNULL( "INV1"."UomEntry", "RDR1"."UomEntry" ) = "PMX_OITM_MEASUREMENTS"."UomEntry"
 LEFT JOIN "OWGT" ON "OWGT"."UnitCode" = "PMX_OITM_MEASUREMENTS"."SWght1Unit"
 WHERE "PMX_PLLI".DocEntry = 3
 GROUP BY "PMX_PLLI"."DocEntry"
				 */
				DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();




				strBuilder.Append("\r\n SELECT ");
				strBuilder.AppendFormat(" ISNULL( SUM( {0} * {1} * 1E-6 * {2} ), 0) ",  // mg -> kg = * 1E-6
					SqlCommandHelper.EscapeTableColumnName(PmxItemMeasurementViewDefinition.NAME, PmxItemMeasurementViewDefinition.Columns.SalesWeight1.NAME),
					SqlCommandHelper.EscapeTableColumnName(SboWeightUnitTable.NAME, SboWeightUnitTable.Columns.WeightInMG.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.Quantity.NAME));



				strBuilder.Append("\r\n FROM ", strBuilder.AsIdentifier(PmxPickListLineTable.NAME));


				strBuilder.Append("\r\n INNER JOIN ", strBuilder.AsIdentifier(PmxItemTable.NAME));
				strBuilder.AppendFormat(" ON {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.ItemCode.NAME));


				strBuilder.Append("\r\n INNER JOIN ", strBuilder.AsIdentifier(PmxPickListProposalLineTable.NAME));
				strBuilder.Append(" ON ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.DocEntry.NAME);
				strBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.BaseEntry.NAME);
				strBuilder.Append(" AND ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.LineNum.NAME);
				strBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.BaseLine.NAME);
				strBuilder.Append(" AND ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.BaseType.NAME);
				strBuilder.Append(" = ", strBuilder.AsString(Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.PmxObjectTypes.PmxPickListProposal));

				strBuilder.Append("\r\n LEFT JOIN ", strBuilder.AsIdentifier(PmxSalesOrderLineTable.NAME));
				strBuilder.Append(" ON ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxSalesOrderLineTable.NAME, PmxSalesOrderLineTable.Columns.DocEntry.NAME);
				strBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseEntry.NAME);
				strBuilder.Append(" AND ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxSalesOrderLineTable.NAME, PmxSalesOrderLineTable.Columns.LineNum.NAME);
				strBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseLine.NAME);
				strBuilder.Append(" AND ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxSalesOrderLineTable.NAME, PmxSalesOrderLineTable.Columns.ObjectType.NAME);
				strBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseType.NAME);



				strBuilder.Append("\r\n LEFT JOIN ", strBuilder.AsIdentifier(PmxSalesInvoiceLineTable.NAME));
				strBuilder.Append(" ON ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxSalesInvoiceLineTable.NAME, PmxSalesInvoiceLineTable.Columns.DocEntry.NAME);
				strBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseEntry.NAME);
				strBuilder.Append(" AND ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxSalesInvoiceLineTable.NAME, PmxSalesInvoiceLineTable.Columns.LineNum.NAME);
				strBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseLine.NAME);
				strBuilder.Append(" AND ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxSalesInvoiceLineTable.NAME, PmxSalesInvoiceLineTable.Columns.ObjectType.NAME);
				strBuilder.Append(" = ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListProposalLineTable.NAME, PmxPickListProposalLineTable.Columns.BaseType.NAME);


				strBuilder.Append("\r\n LEFT JOIN ", strBuilder.AsIdentifier(PmxItemMeasurementViewDefinition.NAME));
				strBuilder.AppendFormat(" ON {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(PmxItemMeasurementViewDefinition.NAME, PmxItemMeasurementViewDefinition.Columns.ItemCode.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxItemTable.NAME, PmxItemTable.Columns.ItemCode.NAME));
				strBuilder.AppendFormat(" AND " + strBuilder.IFNULL + "( {0}, {1} ) = {2}",
					SqlCommandHelper.EscapeTableColumnName(PmxSalesInvoiceLineTable.NAME, PmxSalesInvoiceLineTable.Columns.UomEntry.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxSalesOrderLineTable.NAME, PmxSalesOrderLineTable.Columns.UomEntry.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxItemMeasurementViewDefinition.NAME, PmxItemMeasurementViewDefinition.Columns.UomEntry.NAME));

				strBuilder.Append("\r\n LEFT JOIN ", strBuilder.AsIdentifier(SboWeightUnitTable.NAME));
				strBuilder.AppendFormat(" ON {0} = {1}",
					SqlCommandHelper.EscapeTableColumnName(SboWeightUnitTable.NAME, SboWeightUnitTable.Columns.UnitCode.NAME),
					SqlCommandHelper.EscapeTableColumnName(PmxItemMeasurementViewDefinition.NAME, PmxItemMeasurementViewDefinition.Columns.SalesWght1Unit.NAME));

				strBuilder.Append("\r\n WHERE ");
				SqlCommandHelper.AddTableColumnName(strBuilder, PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.DocEntry.NAME);
				strBuilder.Append(" = ", strBuilder.AsOwnType(pickListDocEntry));

				strBuilder.AppendFormat("\r\n GROUP BY {0}",
					SqlCommandHelper.EscapeTableColumnName(PmxPickListLineTable.NAME, PmxPickListLineTable.Columns.DocEntry.NAME));

				return strBuilder.ToString();
			}

            public static string GetPicklistType( int pickListDocEntry, CultureInfo defaultCultureInfo )
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.Append( " SELECT " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.PickListType.NAME, PmxPickListTable.Columns.PickListType.NAME );
                strBuilder.AppendFormat( " FROM {0} ", SqlCommandHelper.EscapeTableName( PmxPickListTable.NAME ) );
                strBuilder.Append( " WHERE " );
                SqlCommandHelper.AddTableColumnName( strBuilder, PmxPickListTable.NAME, PmxPickListTable.Columns.DocEntry.NAME );
                strBuilder.Append( " = " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( pickListDocEntry, defaultCultureInfo ) );

                return strBuilder.ToString();
            }
        }

        #endregion
    }
}
