﻿//css_ref Produmex.Foundation.SlimScreen.Interfaces;
//css_ref Produmex.Foundation.Wwf.Sbo;
//css_ref Produmex.Foundation.Data.dll;
//css_ref Produmex.Foundation.Data.Sbo.dll;
//css_ref Produmex.Foundation.dll;
//css_ref Produmex.Foundation.Sbo.dll;
//css_ref Produmex.Sbo.Logex.Data.dll;
//css_ref Produmex.Sbo.Logex.Gymna.dll;


using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Common;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Text;
using Produmex.Foundation.Convertors;
using Produmex.Foundation.Data;
using Produmex.Foundation.Data.Sbo;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions.Tables;
using Produmex.Foundation.Data.Sbo.DataObjects;
using Produmex.Foundation.Data.Sbo.Providers;
using Produmex.Foundation.Data.SqlClient;
using Produmex.Foundation.Diagnostics;
using Produmex.Foundation.Messages;
using Produmex.Foundation.SlimScreen;
using Produmex.Foundation.Utilities;
using Produmex.Foundation.Workflows;
using Produmex.Foundation.Workflows.Parameters;
using Produmex.Foundation.Wwf.Sbo.LocalServices;
using Produmex.Sbo.Logex.Data.BusinessObjects;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.Providers;
using Produmex.Foundation;
using Produmex.Foundation.Data.Sbo.BusinessObjects;
using Produmex.Foundation.Data.Sbo.Utilities;
using System.Xml;
using Produmex.Sbo.Logex.Gymna.BusinessObjects.Definitions;
using System.Collections.ObjectModel;
using Produmex.Foundation.Data.DbClient;
using Produmex.Sbo.Logex.Gymna.BusinessObjects.Definitions.Tables;

// PmxWorkflowExecutionTypes.HOOK_FLOW
namespace WorkflowScript_SalesDeliveryCreatedHookScript
{
    /// <summary>
    /// This is a hook script, meant to be customized on a "per project" basis
    /// </summary>
    public class WorkflowScript_SalesDeliveryCreatedHookScript : WorkflowInstanceScriptBase
    {
        private static readonly ILog s_log = LogProvider.GetLogger( MethodInfo.GetCurrentMethod().DeclaringType );


        // Input parameters * do not change *
        public ReadOnlyBinder<CultureInfo> DefaultCultureInfo;
        public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;
        public ReadOnlyBinder<DocumentRef[]> DocumentRefs;

        // Output parameters
        // <none>

        // Sub-flows
        // <none>

        public const string U_PMX_GYM_WEBEMAIL = "U_PMX_GYM_WEBEMAIL";
        public const string U_PMX_G_COLLI = "U_PMX_G_COLLI";
        public const string U_PMX_G_WEIGHT = "U_PMX_G_WEIGHT";
        public const string U_PMX_PLTY = "U_PMX_PLTY";
        public const string PICKLISTTYPE = "GLS";

        /// <summary>
        /// Initializes a new instance of the <see cref="WorkflowScript_SalesDeliveryCreatedHookScript"/> class.
        /// </summary>
        /// <param name="parent">The parent.</param>
        /// <param name="factory">The factory.</param>
        public WorkflowScript_SalesDeliveryCreatedHookScript( WorkflowInstanceBase parent, WorkflowInstanceFactory factory )
            : base( parent, factory )
        {
        }

        #region WorkflowInstanceScriptBase Members

        protected override void Execute()
        {
            // Parameters in scope
            Session session = GetScopeParameter( "Session" ) as Session;
            ISboProviderService sboProviderService = GetScopeParameter( "<WwfService>ISboService" ) as ISboProviderService;

            Message msg = null;

            string query = null;
            string msgKey = null;
            ProdumexError pmxError = null;

            Dictionary<string, PmxParameterValue> tcParameters = null;
            tcParameters = sboProviderService.InvokeMethodWithDbConnection<Dictionary<string, PmxParameterValue>>( false, true, null, null, new DelegateWithPmxDbConnection<Dictionary<string, PmxParameterValue>>( ExecuteGetParameters ), session.DeviceInstanceId );

			string configCodeAccount = null;
			string accountCode = null;
			string packingCode = null;

            //foreach ( DocumentRef docRef in DocumentRefs.Get() )
			if(DocumentRefs.Get().Length > 0)
            {
				DocumentRef docRef = DocumentRefs.Get()[0];

                if (docRef.DocType == SboObjectTypeValues.SalesDelivery
                    || docRef.DocType == SboObjectTypeValues.StockTransfer)
                {
                    bool createGLSFile = sboProviderService.InvokeMethodWithDbConnection<bool>(false, false, null, null,
                        delegate(PmxDbConnection conn, object[] parameters)
                        {
                            switch (docRef.DocType)
                            {
                                case SboObjectTypeValues.SalesDelivery: query = BuildQuery.GetDeliveryData(docRef.DocEntry, DefaultCultureInfo); break;
                                case SboObjectTypeValues.StockTransfer: query = BuildQuery.GetStockTransferData(docRef.DocEntry, conn.DbTool); break;
                            }
                            using (DbDataReader rdr = conn.CreateCommand(query).ExecuteReader())
                            {
                                if (!rdr.Read())
                                {
                                    return false;
                                }

                                string picklistType = rdr[U_PMX_PLTY].ToString();
                                if (picklistType != PICKLISTTYPE)
                                {
                                    s_log.InfoFormat("Unable to create label file for GLS for {0}. Wrong picklist type.", docRef);
                                    return false;
                                }

                                switch (docRef.DocType)
                                {
                                    case SboObjectTypeValues.StockTransfer:
                                        if (rdr.Read())
                                        {
                                            return false;  // More than 1 ship-to found, no label
                                        }
                                        break;
                                }
                            }
                            return true;
                        });


					if (createGLSFile)
					{

						try
						{
							if (string.IsNullOrEmpty(accountCode))
							{
								List<string> buttons = new List<string>();
								buttons.Add("Parcel");
								buttons.Add("Colli");
								buttons.Add("Pallet");

							Step_SelectPackingCode:
								//Get packing code
								session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectButtonScreen),
							   DefaultCultureInfo.Get(), BuildParamCollection(
							   "InitialErrorKey", null,
							   "TitleKey", "Selecteer verpakking",
							   "AllowedButtons", buttons.ToArray()
							   ));
								msg = WaitForMessage();
								if (msg.Name.EndsWith(".ButtonClicked"))
								{
									string filterButtonKey = ExtractParameter<string>(msg.Parameters, "buttonKey");
									if (filterButtonKey == "Parcel")
									{
										configCodeAccount = PmxConfigCodesGymna.GLS_ACCOUNT_NUMBER_PARCEL;
										packingCode = "PCO";
									}
									else if (filterButtonKey == "Colli")
									{
										configCodeAccount = PmxConfigCodesGymna.GLS_ACCOUNT_NUMBER_FREIGHT;
										packingCode = "CO";
									}
									else
									{
										configCodeAccount = PmxConfigCodesGymna.GLS_ACCOUNT_NUMBER_FREIGHT;
										packingCode = "PL";
									}
								}
								else
								{
									//go back is not allowed
									goto Step_SelectPackingCode;
								}

								sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
								   delegate(PmxDbConnection conn, object[] parameters)
								   {
									   PmxConfigProvider cfgProv = new PmxConfigProvider(conn);
									   using (PmxConfig cfg = cfgProv.GetBO(configCodeAccount))
									   {
										   accountCode = cfg.ConfigValue;
									   }

									   return null;
								   });

							}
							sboProviderService.InvokeMethodWithDbConnection<bool>(false, true, null, null, CreateGLSFile,
                                session.DeviceInstanceId, tcParameters, docRef, accountCode, packingCode);
						}
						catch (ProdumexException ex)
						{
							s_log.Error(ex.ToString());
							msgKey = ex.Message;
							pmxError = ex.Error;
						}
						catch (Exception ex)
						{
							s_log.Error("", ex);
							msgKey = ex.Message;
						}

						if (msgKey != null)
						{
							session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
								DefaultCultureInfo.Get(), BuildParamCollection(
									"InitialErrorKey", null,
									"MessageKey", msgKey,
									"PmxError", pmxError,
									"ShowButton", true));
							WaitForMessage();
						}


                        System.Threading.Thread.Sleep(500);  // 2016-02-18  KrisM  What problem? Can this be moved inside the if (createGLSFile) ?
                    }
                }
            }
            //wait, to avoid problems on wave
            // System.Threading.Thread.Sleep(500);  // 2016-02-18  KrisM  What problem? Can this be moved inside the if (createGLSFile) ?

            return;
        }

        #endregion



        #region HELPERS






        /// <summary>
        /// Gets the number of rows.
        /// </summary>
        /// <param name="ds">The ds.</param>
        /// <returns></returns>
        private int GetNumberOfRows(DataSet ds)
        {
            if (ds != null && ds.Tables.Count > 0)
            {
                return ds.Tables[0].Rows.Count;
            }
            return 0;
        }

        /// <summary>
        /// Gets the parameters.
        /// </summary>
        /// <param name="dbConn">The db conn.</param>
        /// <param name="parameters">The parameters.</param>
        /// <returns></returns>
        private Dictionary<string, PmxParameterValue> ExecuteGetParameters( PmxDbConnection dbConn, object[] parameters )
        {
            string deviceInstanceId = (string)parameters[0];

            PmxParameterValueProvider parProv = new PmxParameterValueProvider( dbConn );
            return parProv.GetParameterValuesAsDictionary( String.Format( "{0}->{1}", deviceInstanceId, PmxPropertyCodes.OSE_PARAMETERSET ), true, false );

        }



        /// <summary>
        /// Creates the GLS file.
        /// </summary>
        /// <param name="dbConn">The database connection.</param>
        /// <param name="parameters">The parameters.</param>
        /// <returns></returns>
        private bool CreateGLSFile(PmxDbConnection dbConn, object[] parameters )
        {
            Dictionary<string, PmxParameterValue> tcParameters = (Dictionary<string, PmxParameterValue>)parameters[1];
            DocumentRef docRef = (DocumentRef)parameters[2];
            string accountCode = (string)parameters[3];
			string packingCode = (string)parameters[4];

            string deliveryNumber = null;
            StringBuilder sbDoc = new StringBuilder();

            string query = null;
            switch (docRef.DocType)
            {
                case SboObjectTypeValues.SalesDelivery: query = BuildQuery.GetDeliveryData(docRef.DocEntry, DefaultCultureInfo); break;
                case SboObjectTypeValues.StockTransfer: query = BuildQuery.GetStockTransferData(docRef.DocEntry, dbConn.DbTool); break;
            }
            using ( ISboRecordset rs = SboRecordsetHelper.RunQuery( s_log, query, dbConn ) )
            {
                if (rs.EoF)
                {
                    s_log.WarnFormat("Unable to create label file for GLS for {0}. Missing data.", docRef);
                    return false;
                }
				/*
				A - GLS Customer number: Always 100730540 fro Gymna. Values actually stored in config table in DB
				B - Customer name 1: ODLN.CardName -> Replaced with ODLN.ShipToCode by FR on 2016-01-04 
				C - Customer name 2: Empty
				D - Street, number, ...: Format: DLN12.StreetS, DLN12.BuildingS
				E - n/A
				F - n/a
				G - DLN12.ZipCodeS
				H - DLN12.CityS
				I - DLN12.CountryS
				J - Phonenumber: OCRD.Phone1
				K - n/A
				L - n/a
				M - Reference: ODLN.NumAtCard
				N - Unit: Number of colli: ODLN_U_PMX_G_COLLI
				O - Weight: ODLN.U_PMX_G_WEIGHT
				P - PackagingCode: User selected option: Parcel : PCO, Freight grote colli  : CO, Freight op pallet : PL
				Q - E mail: OCPR.E_Mail
				 */
                deliveryNumber = rs.GetTypedValue<int>( SboSalesDeliveryNoteTable.Columns.DocNum.NAME ).ToString();
                //string customerName1 = rs.GetTypedValue<string>( SboSalesDeliveryNoteTable.Columns.CardName.NAME );
				string customerName1 = rs.GetTypedValue<string>(SboSalesDeliveryNoteTable.Columns.ShipToCode.NAME);
				string customerName2 = "";
                string street = rs.GetTypedValue<string>( SboSalesDeliveryTaxExtensionTable.Columns.StreetS.NAME );
                string building = rs.GetTypedValue<string>( SboSalesDeliveryTaxExtensionTable.Columns.BuildingS.NAME );
                string zipCode = rs.GetTypedValue<string>( SboSalesInvoiceTaxExtensionTable.Columns.ZipCodeS.NAME );
                string city = rs.GetTypedValue<string>( SboSalesInvoiceTaxExtensionTable.Columns.CityS.NAME );
                string countryCode = rs.GetTypedValue<string>( SboSalesInvoiceTaxExtensionTable.Columns.CountryS.NAME );
                string phoneNumber = rs.GetTypedValue<string>( SboBusinessPartnerTable.Columns.Phone1.NAME ) == null ? null : rs.GetTypedValue<string>( SboBusinessPartnerTable.Columns.Phone1.NAME ).Replace( "/", "" );
                string reference = rs.GetTypedValue<int>( SboSalesDeliveryNoteTable.Columns.DocNum.NAME ).ToString();
                string unit = rs.GetTypedValue<int>( U_PMX_G_COLLI ).ToString();
				//Weight in KG and 2 decimals, without decimal sign
				string weight = rs.GetTypedValue<double>(U_PMX_G_WEIGHT).ToString("#.00");
                string gymEmail = rs.GetTypedValue<string>(U_PMX_GYM_WEBEMAIL);
                string email = string.IsNullOrEmpty(gymEmail) ? rs.GetTypedValue<string>(SboContactPersonTable.Columns.EmailAddress.NAME) : gymEmail;

                if ( customerName1 != null && street != null && countryCode != null && zipCode != null && city != null)
                {
                    //5., 6., 11., 12. columns are empty - not "n/a"
                    sbDoc.AppendFormat( "{0};{1};{2};{3},{4};{5};{6};{7};{8};{9};{10};{11};{12};{13};{14};{15};{16};{17}",
                        accountCode, customerName1, customerName2, street, building, "", "", zipCode, city, countryCode, phoneNumber,
                        "", "", reference, unit, weight, packingCode, email );
                }
                else
                {
                    s_log.WarnFormat("Unable to create label file for GLS for {0}. Missing data .", docRef);
					return false;
                }
            }

            if ( string.IsNullOrEmpty( deliveryNumber ) )
            {
                s_log.WarnFormat("Unable to create label file for GLS for {0}. No delivery was found.", docRef);
                return false;
            }
                
            // Write file
            string outputDir = PmxParameterValueProvider.GetParameterValue<string>( tcParameters, PmxExtensionParameterCodesG.GLS_OUT_DIR );

            // Make sure path ends with back-slash
            if ( !outputDir.EndsWith( @"\" ) )
            {
                outputDir += @"\";
            }

            string outputDirTemp = Path.Combine(outputDir, "Temp");

            System.IO.Directory.CreateDirectory( outputDir );
            System.IO.Directory.CreateDirectory( outputDirTemp );


            string filename = string.Format("{0}_{1}.csv", DateTime.Now.ToString("yyyyMMddHHmmssFFF"), docRef.DocEntry);


            using ( StreamWriter writer = new StreamWriter( Path.Combine( outputDirTemp, filename ), false, Encoding.GetEncoding( 1252 ) ) )
            {
                writer.Write( sbDoc.ToString() );
                writer.Flush();
            }

            //if file exists in destination, we remove it
            if ( File.Exists( System.IO.Path.Combine( outputDir, filename ) ) )
            {
                File.Delete( System.IO.Path.Combine( outputDir, filename ) );
            }

            //move created file to correct folder
            System.IO.File.Move( System.IO.Path.Combine( outputDirTemp, filename ), System.IO.Path.Combine( outputDir, filename ) );


            return true;
        }

        #endregion

        #region QUERY

        private class BuildQuery
        {

            public static string GetDeliveryData( int docEntry, CultureInfo defaultCultureInfo )
            {
                /*
                 select 
                 ODLN.CardName,
                 DLN12.StreetS,
                 DLN12.BuildingS,
                 DLN12.ZipCodeS,
                 DLN12.CityS,
                 DLN12.CountryS,
                 OCRD.Phone1,
                 ODLN.NumAtCard,
                 ODLN.U_PMX_G_COLLI,
                 ODLN.U_PMX_G_WEIGHT,
                 OCPR.E_MailL
                 from ODLN
                 inner join DLN12 on ODLN.DocEntry = DLN12.DocEntry
                 left join OCPR on odln.CntctCode = OCPR.CntctCode
                 where ODLN.DocEntry = {}
                 */
                StringBuilder strBuilder = new StringBuilder(); 

                strBuilder.Append( " SELECT " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.DocNum.NAME, SboSalesDeliveryNoteTable.Columns.DocNum.NAME );
                strBuilder.Append( ",\r\n " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.CardName.NAME, SboSalesDeliveryNoteTable.Columns.CardName.NAME );
                strBuilder.Append( ",\r\n " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboSalesDeliveryTaxExtensionTable.NAME, SboSalesDeliveryTaxExtensionTable.Columns.StreetS.NAME, SboSalesDeliveryTaxExtensionTable.Columns.StreetS.NAME );
                strBuilder.Append( ",\r\n " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboSalesDeliveryTaxExtensionTable.NAME, SboSalesDeliveryTaxExtensionTable.Columns.BuildingS.NAME, SboSalesDeliveryTaxExtensionTable.Columns.BuildingS.NAME );
                strBuilder.Append( ",\r\n " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboSalesDeliveryTaxExtensionTable.NAME, SboSalesDeliveryTaxExtensionTable.Columns.ZipCodeS.NAME, SboSalesDeliveryTaxExtensionTable.Columns.ZipCodeS.NAME );
                strBuilder.Append( ",\r\n " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboSalesDeliveryTaxExtensionTable.NAME, SboSalesDeliveryTaxExtensionTable.Columns.CityS.NAME, SboSalesDeliveryTaxExtensionTable.Columns.CityS.NAME );
                strBuilder.Append( ",\r\n " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboSalesDeliveryTaxExtensionTable.NAME, SboSalesDeliveryTaxExtensionTable.Columns.CountryS.NAME, SboSalesDeliveryTaxExtensionTable.Columns.CountryS.NAME );
                strBuilder.Append( ",\r\n " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboBusinessPartnerTable.NAME, SboBusinessPartnerTable.Columns.Phone1.NAME, SboBusinessPartnerTable.Columns.Phone1.NAME );
                strBuilder.Append( ",\r\n " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.NumAtCard.NAME, SboSalesDeliveryNoteTable.Columns.NumAtCard.NAME );
                strBuilder.Append( ",\r\n " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboSalesDeliveryNoteTable.NAME, U_PMX_G_COLLI, U_PMX_G_COLLI );
                strBuilder.Append( ",\r\n " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboSalesDeliveryNoteTable.NAME, U_PMX_G_WEIGHT, U_PMX_G_WEIGHT );
                strBuilder.Append( ",\r\n " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboContactPersonTable.NAME, SboContactPersonTable.Columns.EmailAddress.NAME, SboContactPersonTable.Columns.EmailAddress.NAME );
                strBuilder.Append( ",\r\n " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboSalesDeliveryNoteTable.NAME, U_PMX_PLTY, U_PMX_PLTY );
                strBuilder.Append(",\r\n ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, SboSalesDeliveryNoteTable.NAME, U_PMX_GYM_WEBEMAIL, U_PMX_GYM_WEBEMAIL);
                strBuilder.Append( ",\r\n " );
                SqlCommandHelper.AddTableColumnNameWithAs( strBuilder, SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.ShipToCode.NAME, SboSalesDeliveryNoteTable.Columns.ShipToCode.NAME );
                
				strBuilder.AppendFormat( " FROM {0} ", SqlCommandHelper.EscapeTableName(SboSalesDeliveryNoteTable.NAME) );
                strBuilder.AppendFormat( " INNER JOIN {0} ON {1} = {2} ", SqlCommandHelper.EscapeTableName( SboSalesDeliveryTaxExtensionTable.NAME ),
                                                                          SqlCommandHelper.EscapeTableColumnName( SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.DocEntry.NAME ),
                                                                          SqlCommandHelper.EscapeTableColumnName( SboSalesDeliveryTaxExtensionTable.NAME, SboSalesDeliveryTaxExtensionTable.Columns.DocEntry.NAME ) );

                strBuilder.AppendFormat( " LEFT JOIN {0} ON {1} = {2} ", SqlCommandHelper.EscapeTableName( SboContactPersonTable.NAME ),
                     SqlCommandHelper.EscapeTableColumnName( SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.CntctCode.NAME ),
                     SqlCommandHelper.EscapeTableColumnName( SboContactPersonTable.NAME, SboContactPersonTable.Columns.CntctCode.NAME ) );
                strBuilder.AppendFormat( " LEFT JOIN {0} ON {1} = {2} ", SqlCommandHelper.EscapeTableName( SboBusinessPartnerTable.NAME ),
                     SqlCommandHelper.EscapeTableColumnName( SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.CardCode.NAME ),
                     SqlCommandHelper.EscapeTableColumnName( SboBusinessPartnerTable.NAME, SboBusinessPartnerTable.Columns.CardCode.NAME ) );
                strBuilder.Append( " WHERE " );
                SqlCommandHelper.AddTableColumnName( strBuilder, SboSalesDeliveryNoteTable.NAME, SboSalesDeliveryNoteTable.Columns.DocEntry.NAME );
                strBuilder.Append( " = " );
                strBuilder.Append( SqlCommandHelper.FormatAndEscapeColumnValue( docEntry, defaultCultureInfo ) );

                return strBuilder.ToString();
            }


            internal static string GetStockTransferData(int docEntry, DbTool dbTool)
            {
                DbQueryBuilder qb = dbTool.GetQueryBuilder();

                qb.Append(
                    "SELECT ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxStockTransferRequestTable.Columns.PickListType.NAME),  // From PmxStockTransferTable

                         ", ", qb.AsIdentifierWithAlias(PmxStockTransferTable.NAME, PmxStockTransferTable.Columns.DocNum.NAME, SboSalesDeliveryNoteTable.Columns.DocNum.NAME),
                         ", COALESCE( ", qb.AsIdentifier(SboWarehouseTable.NAME, SboWarehouseTable.Columns.WarehouseName.NAME),
                                   ", ", qb.AsIdentifier(SboSalesOrderTable.NAME, SboSalesOrderTable.Columns.CardName.NAME),
                                   ", ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxStockTransferTable.Columns.CardName.NAME),  // Last because always non-null
                         " ) AS ", qb.AsIdentifier(PmxStockTransferTable.Columns.CardName.NAME),
                         ", COALESCE( ", qb.AsIdentifier(SboWarehouseTable.NAME, SboWarehouseTable.Columns.WarehouseName.NAME),
                                   ", ", qb.AsIdentifier(SboSalesOrderTable.NAME, SboSalesOrderTable.Columns.ShipToCode.NAME),
                                   ", ", qb.AsIdentifier(SboBusinessPartnerAddressTable.NAME, SboBusinessPartnerAddressTable.Columns.Address.NAME),  // Last because always non-null
                         " ) AS ", qb.AsIdentifier(SboSalesDeliveryNoteTable.Columns.ShipToCode.NAME),
                         ", ", qb.AsIdentifier(SboSalesOrderTaxExtensionTable.NAME, SboSalesOrderTaxExtensionTable.Columns.BuildingS.NAME),
                         ", COALESCE( ", qb.AsIdentifier(SboWarehouseTable.NAME, SboWarehouseTable.Columns.Street.NAME),
                                   ", ", qb.AsIdentifier(SboBusinessPartnerAddressTable.NAME, SboBusinessPartnerAddressTable.Columns.Street.NAME),
                                   ", ", qb.AsIdentifier(SboSalesOrderTaxExtensionTable.NAME, SboSalesOrderTaxExtensionTable.Columns.StreetS.NAME),
                         " ) AS ", qb.AsIdentifier(SboSalesDeliveryTaxExtensionTable.Columns.StreetS.NAME),
                         ", COALESCE( ", qb.AsIdentifier(SboWarehouseTable.NAME, SboWarehouseTable.Columns.ZipCode.NAME),
                                   ", ", qb.AsIdentifier(SboBusinessPartnerAddressTable.NAME, SboBusinessPartnerAddressTable.Columns.ZipCode.NAME),
                                   ", ", qb.AsIdentifier(SboSalesOrderTaxExtensionTable.NAME, SboSalesOrderTaxExtensionTable.Columns.ZipCodeS.NAME),
                         " ) AS ", qb.AsIdentifier(SboSalesDeliveryTaxExtensionTable.Columns.ZipCodeS.NAME),
                         ", COALESCE( ", qb.AsIdentifier(SboWarehouseTable.NAME, SboWarehouseTable.Columns.City.NAME),
                                   ", ", qb.AsIdentifier(SboBusinessPartnerAddressTable.NAME, SboBusinessPartnerAddressTable.Columns.City.NAME),
                                   ", ", qb.AsIdentifier(SboSalesOrderTaxExtensionTable.NAME, SboSalesOrderTaxExtensionTable.Columns.CityS.NAME),
                         " ) AS ", qb.AsIdentifier(SboSalesDeliveryTaxExtensionTable.Columns.CityS.NAME),
                         ", COALESCE( ", qb.AsIdentifier(SboWarehouseTable.NAME, SboWarehouseTable.Columns.Country.NAME),
                                   ", ", qb.AsIdentifier(SboBusinessPartnerAddressTable.NAME, SboBusinessPartnerAddressTable.Columns.Country.NAME),
                                   ", ", qb.AsIdentifier(SboSalesOrderTaxExtensionTable.NAME, SboSalesOrderTaxExtensionTable.Columns.CountryS.NAME),
                         " ) AS ", qb.AsIdentifier(SboSalesDeliveryTaxExtensionTable.Columns.CountryS.NAME),
                         ", ", qb.AsIdentifier(SboBusinessPartnerTable.NAME, SboBusinessPartnerTable.Columns.Phone1.NAME),
                         ", ", qb.AsIdentifier(SboContactPersonTable.NAME, SboContactPersonTable.Columns.EmailAddress.NAME),

                         ", ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxSalesDeliveryNoteGTable.Columns.NumberOfColli.NAME),
                         ", ", qb.AsIdentifier(PmxSalesOrderTable.NAME, U_PMX_GYM_WEBEMAIL),
                         ", ", qb.AsIdentifier(PmxStockTransferTable.NAME, U_PMX_G_WEIGHT),


                "\r\n FROM ", qb.AsIdentifier(PmxStockTransferTable.NAME),

                    // Transfer to non-KLA warehouse, send to referenced warehouse
                    "\r\n LEFT JOIN ", qb.AsIdentifier(PmxWarehouseTable.NAME),
                               " ON ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxStockTransferTable.Columns.ToWarehouse.NAME),
                               " <> ", qb.AsString(Defs.WarehouseKLA),
                              " AND ", qb.AsIdentifier(PmxWarehouseTable.NAME, PmxWarehouseTable.Columns.WarehouseCode.NAME),
                                " = ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxStockTransferTable.Columns.ToWarehouse.NAME),

                    // Transfer to KLA warehouse, send to customer if no SalesOrderDocNum
                    "\r\n LEFT JOIN ", qb.AsIdentifier(SboBusinessPartnerAddressTable.NAME),
                               " ON ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxStockTransferTable.Columns.ToWarehouse.NAME),
                                " = ", qb.AsString(Defs.WarehouseKLA),
                              " AND ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxStockTransferRequestTableG.Columns.SalesOrderDocNum.NAME),
                                       " IS NULL",
                              " AND ", qb.AsIdentifier(SboBusinessPartnerAddressTable.NAME, SboBusinessPartnerAddressTable.Columns.CardCode.NAME),
                                " = ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxStockTransferTable.Columns.CardCode.NAME),
                              " AND ", qb.AsIdentifier(SboBusinessPartnerAddressTable.NAME, SboBusinessPartnerAddressTable.Columns.Type.NAME),
                                " = ", qb.AsString(SboBusinessPartnerAddressTypeColumnValues.ShipTo.VALUE),

                    // Transfer to KLA warehouse, send to SalesOrder address if SalesOrderDocNum available
                    "\r\n LEFT JOIN ( ", qb.AsIdentifier(PmxSalesOrderTable.NAME),
                    "\r\n    INNER JOIN ", qb.AsIdentifier(SboSalesOrderTaxExtensionTable.NAME),
                                   " ON ", qb.AsIdentifier(SboSalesOrderTaxExtensionTable.NAME, SboSalesOrderTaxExtensionTable.Columns.DocEntry.NAME),
                                    " = ", qb.AsIdentifier(PmxSalesOrderTable.NAME, PmxSalesOrderTable.Columns.DocEntry.NAME),
                    "\r\n ) ON ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxStockTransferTable.Columns.ToWarehouse.NAME),
                           " = ", qb.AsString(Defs.WarehouseKLA),
                         " AND ", qb.AsIdentifier(PmxSalesOrderTable.NAME, PmxSalesOrderTable.Columns.DocNum.NAME),
                           " = ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxStockTransferRequestTableG.Columns.SalesOrderDocNum.NAME),


                    "\r\n LEFT JOIN ", qb.AsIdentifier(SboBusinessPartnerTable.NAME),
                                " ON ", qb.AsIdentifier(SboBusinessPartnerTable.NAME, SboBusinessPartnerTable.Columns.CardCode.NAME),
                                " = COALESCE( ", qb.AsIdentifier(PmxSalesOrderTable.NAME, PmxSalesOrderTable.Columns.CardCode.NAME),
                                           ", ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxStockTransferTable.Columns.CardCode.NAME), " )",

                    "\r\n LEFT JOIN ", qb.AsIdentifier(SboContactPersonTable.NAME),
                                " ON ", qb.AsIdentifier(SboContactPersonTable.NAME, SboContactPersonTable.Columns.CntctCode.NAME),
                                " = COALESCE( ", qb.AsIdentifier(PmxSalesOrderTable.NAME, PmxSalesOrderTable.Columns.CntctCode.NAME),
                                           //", ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxStockTransferTable.Columns.CntctCode.NAME), " )",
											", ", qb.AsIdentifier(PmxStockTransferTable.NAME, "CntctCode"), " )",

                    "\r\n WHERE ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxStockTransferTable.Columns.DocEntry.NAME), " = ", docEntry,
                          " AND ", qb.AsIdentifier(PmxStockTransferTable.NAME, PmxStockTransferTable.Columns.FromWareHouse.NAME), " = ", qb.AsString(Defs.WarehouseB10),
                          " AND ( ", qb.AsIdentifier(PmxWarehouseTable.NAME, PmxWarehouseTable.Columns.WarehouseCode.NAME), " IS NOT NULL",
                             " OR ", qb.AsIdentifier(SboBusinessPartnerAddressTable.NAME, SboBusinessPartnerAddressTable.Columns.CardCode.NAME), " IS NOT NULL",
                             " OR ", qb.AsIdentifier(PmxSalesOrderTable.NAME, PmxSalesOrderTable.Columns.DocNum.NAME), " IS NOT NULL )");

                return qb.ToString();
            }
        }

        private struct Defs
        {
            public const string WarehouseB10 = "B10";
            public const string WarehouseKLA = "KLA";
        }

        #endregion
    }
}
